"""Handshake protocol."""

from __future__ import annotations
from typing import NamedTuple

from mcipc.query.proto.common import MAGIC, random_session_id, Type


__all__ = ['Request', 'Response', 'HandshakeMixin']


class Request(NamedTuple):
    """A client → server handshake request packet."""

    magic: bytes
    type: Type
    session_id: int

    def __bytes__(self):
        """Converts the packet to bytes."""
        payload = self.magic
        payload += bytes(self.type)
        payload += self.session_id.to_bytes(4, 'big', signed=True)
        return payload

    @classmethod
    def create(cls, session_id: int = None) -> Request:
        """Returns a handshake request packet with a random session ID."""
        if session_id is None:
            session_id = random_session_id()

        return cls(MAGIC, Type.HANDSHAKE, session_id)


class Response(NamedTuple):
    """A server → client handshake response packet."""

    type: bytes
    session_id: int
    challenge_token: int

    @classmethod
    def from_bytes(cls, bytes_: bytes) -> Response:
        """Creates the packet from bytes."""
        type_ = Type.from_bytes(bytes_[0:1])
        session_id = int.from_bytes(bytes_[1:5], 'big', signed=True)
        challenge_token = bytes_[5:-1].decode()
        return cls(type_, session_id, int(challenge_token))

    def to_json(self) -> dict:
        """Returns a JSON-ish dict."""
        return {
            'type': self.type.value,
            'session_id': self.session_id,
            'challenge_token': self.challenge_token
        }


class HandshakeMixin:   # pylint: disable=R0903
    """Query client mixin for performing handshakes."""

    def handshake(self) -> Response:
        """Performs a handshake."""
        request = Request.create()
        return self.communicate(request, Response)



"""Information about online players."""

from __future__ import annotations
from re import Match, compile  # pylint: disable=W0622
from typing import NamedTuple, Tuple, Union


__all__ = ['Players']


REGEX_VANILLA = compile('.+ (\\d+) .+ (\\d+) .+: (.*)')
REGEX_PAPER = compile('.+ §c(\\d+)§6 .+ §c(\\d+)§6 .+\\.([\\s\\S]*)')
REGEX_PAPER_NAME = compile('.+: (?:§4)?(\\w+)(?:§r)?§f')


def extract_paper_name(name: str) -> Union[str, None]:
    """Extracts names from Paper server output."""

    match = REGEX_PAPER_NAME.fullmatch(name.strip())

    if match is None:
        return None

    name, = match.groups()
    return name


class Players(NamedTuple):
    """Online players information."""

    online: int
    max: int
    names: Tuple[str]

    @classmethod
    def from_vanilla(cls, match: Match) -> Players:
        """Creates the players information from a vanilla server match."""
        online, max_, names = match.groups()
        names = filter(None, map(lambda name: name.strip(), names.split(', ')))
        return cls(int(online), int(max_), tuple(names))

    @classmethod
    def from_paper(cls, match: Match) -> Players:
        """Creates the players information from a Papers server response.
        https://github.com/conqp/mcipc/issues/13#issuecomment-726145034
        """
        online, max_, names = match.groups()
        names = filter(None, map(extract_paper_name, names.split('\n')))
        return cls(int(online), int(max_), tuple(names))

    @classmethod
    def from_response(cls, text: str) -> Players:
        """Creates the players information from a server response."""
        match = REGEX_VANILLA.fullmatch(text)

        if match is not None:
            return cls.from_vanilla(match)

        match = REGEX_PAPER.fullmatch(text)

        if match is not None:
            return cls.from_paper(match)

        raise ValueError('Unexpected players response:', text)

    def to_json(self) -> dict:
        """Returns a JSON-ish dict."""
        return {'online': self.online, 'max': self.max, 'names': self.names}



"""Common enumerations."""

from __future__ import annotations
from enum import Enum
from typing import Callable

from mcipc.server.datatypes import VarInt


__all__ = ['State']


class State(Enum):
    """Protocol state."""

    HANDSHAKE = VarInt(0)
    STATUS = VarInt(1)
    LOGIN = VarInt(2)

    @classmethod
    def read(cls, readfunc: Callable) -> State:
        """Reads the state from the respective connection."""
        return cls(VarInt.read(readfunc))



"""Data structures to represent server information."""

from mcipc.rcon.datastructures.help import Help
from mcipc.rcon.datastructures.location import Location
from mcipc.rcon.datastructures.players import Players
from mcipc.rcon.datastructures.seed import Seed


__all__ = ['Help', 'Location', 'Players', 'Seed']



"""Server configuration parser."""

from __future__ import annotations
from configparser import ConfigParser
from pathlib import Path
from typing import NamedTuple


__all__ = [
    'LOG_FORMAT',
    'FORTUNE',
    'InvalidCredentials',
    'CredentialsConfig'
]


LOG_FORMAT = '[%(levelname)s] %(name)s: %(message)s'
FORTUNE = Path('/usr/bin/fortune')


class InvalidCredentials(ValueError):
    """Indicates invalid credentials."""


class CredentialsConfig(ConfigParser):  # pylint: disable=R0901
    """Parses the RCON config file."""

    def __init__(self, filename):
        """Sets the file name."""
        super().__init__()
        self.filename = filename

    @property
    def servers(self) -> dict:
        """Returns a dictionary of servers."""
        self.read(self.filename)
        return {
            section: Credentials.from_config_section(self[section])
            for section in self.sections()
        }


class Credentials(NamedTuple):
    """Represents server credentials."""

    host: str
    port: int
    passwd: str

    @classmethod
    def from_string(cls, string) -> Credentials:
        """Reads the credentials from the given string."""
        try:
            host, port = string.split(':')
        except ValueError:
            raise InvalidCredentials(f'Invalid socket: {string}.') from None

        try:
            port = int(port)
        except ValueError:
            raise InvalidCredentials(f'Not an integer: {port}.') from None

        try:
            passwd, host = host.rsplit('@', maxsplit=1)
        except ValueError:
            passwd = None

        return cls(host, port, passwd)

    @classmethod
    def from_config_section(cls, section) -> Credentials:
        """Creates a credentials tuple from
        the respective config section.
        """
        host = section['host']
        port = section.getint('port')
        passwd = section.get('passwd')
        return cls(host, port, passwd)



"""RCON client library."""

from mcipc.rcon.client import Client
from mcipc.rcon.console import rconcmd
from mcipc.rcon.exceptions import InvalidPacketStructure
from mcipc.rcon.exceptions import StructureNotFound
from mcipc.rcon.exceptions import WrongPassword


__all__ = [
    'InvalidPacketStructure',
    'StructureNotFound',
    'WrongPassword',
    'rconcmd',
    'Client'
]



"""Low-level protocol stuff."""

from typing import Union

from mcipc.query.proto.basic_stats import BasicStatsMixin
from mcipc.query.proto.basic_stats import Request as BasicStatsRequest
from mcipc.query.proto.full_stats import FullStatsMixin
from mcipc.query.proto.full_stats import Request as FullStatsRequest
from mcipc.query.proto.handshake import HandshakeMixin
from mcipc.query.proto.handshake import Request as HandshakeRequest


__all__ = ['BasicStatsMixin', 'FullStatsMixin', 'HandshakeMixin', 'Packet']


Packet = Union[BasicStatsRequest, FullStatsRequest, HandshakeRequest]



"""Represents a seed."""

from __future__ import annotations
from re import compile  # pylint: disable=W0622


__all__ = ['Seed']


REGEX = compile('.*\\[(-?\\d+)\\]')


class Seed(int):
    """A seed value."""

    @classmethod
    def from_response(cls, text: str) -> Seed:
        """Returns a seed from a server response."""
        match = REGEX.fullmatch(text)

        if match is None:
            raise ValueError('Unexpected seed response:', text)

        return cls(match.group(1))

    def to_json(self) -> int:
        """Returns a JSON-ish dict."""
        return int(self)



"""Stuff, common to Query and RCON."""

from socket import socket, SocketKind   # pylint: disable=E0611


__all__ = ['BaseClient']


class BaseClient:
    """A basic client."""

    def __init__(self, typ: SocketKind, host: str, port: int,
                 timeout: float = None):
        """Sets host an port."""
        self._socket = socket(type=typ)
        self.host = host
        self.port = port
        self.timeout = timeout

    def __enter__(self):
        """Conntects the socket."""
        self._socket.__enter__()
        self.connect()
        return self

    def __exit__(self, typ, value, traceback):
        """Delegates to the underlying socket's exit method."""
        self.close()
        return self._socket.__exit__(typ, value, traceback)

    @property
    def socket(self) -> tuple:
        """Returns a tuple of host and port."""
        return (self.host, self.port)

    def connect(self):
        """Conntects to the RCON server."""
        self._socket.settimeout(self.timeout)
        return self._socket.connect(self.socket)

    def close(self):
        """Disconnects from the RCON server."""
        return self._socket.close()



"""More complex data structures."""

from __future__ import annotations
from json import dumps, loads
from logging import getLogger
from typing import Callable, NamedTuple

from mcipc.server.datatypes import VarInt
from mcipc.server.enumerations import State


__all__ = ['Handshake']


LOGGER = getLogger(__file__)


class Handshake(NamedTuple):
    """Server-bound handshake packet."""

    protocol: VarInt
    address: str
    port: int
    next_state: State

    @classmethod
    def read(cls, readfunc: Callable) -> Handshake:
        """Creates a handshake object from the respective bytes."""
        size = VarInt.read(readfunc)
        LOGGER.debug('Read size: %s', size)
        version = VarInt.read(readfunc)
        LOGGER.debug('Read version: %s', version)
        version_length = len(bytes(version))
        payload_size = size - version_length - 1    # Do not read next state.
        payload = readfunc(payload_size)
        LOGGER.debug('Read payload: %s', payload)
        address = payload[4:-2].decode()
        port = int.from_bytes(payload[-2:], 'little')
        next_state = State.read(readfunc)
        return cls(version, address, port, next_state)


class SLPResponse(NamedTuple):
    """A server list ping response."""

    packet_id: VarInt
    json: dict

    def __bytes__(self):
        """Returns the respective bytes."""
        json = dumps(self).encode('latin-1')
        json_size = len(json)
        json_size = VarInt(json_size)
        json_size = bytes(json_size)
        payload = json_size + json
        payload = bytes(self.packet_id) + payload
        payload_size = len(payload)
        payload_size = VarInt(payload_size)
        payload_size = bytes(payload_size)
        return payload_size + payload

    @classmethod
    def read(cls, readfunc: Callable) -> SLPResponse:
        """Creates the SLP response from the respective payload."""
        total_size = VarInt.read(readfunc)
        LOGGER.debug('Read total size: %s', total_size)
        packet_id = VarInt.read(readfunc)
        LOGGER.debug('Read packet ID: %s', packet_id)
        json_size = VarInt.read(readfunc)
        LOGGER.debug('Read JSON size: %s', json_size)
        json_bytes = readfunc(json_size)
        LOGGER.debug('Read JSON bytes: %s', json_bytes)
        string = json_bytes.decode('latin-1')
        json = loads(string)
        return cls(packet_id, json)



"""Data types for the server protocol."""

from __future__ import annotations
from logging import getLogger
from typing import Callable

from mcipc.server.functions import rshift


__all__ = ['VarInt']


LOGGER = getLogger(__file__)


class VarInt(int):
    """Minecraft protocol VarInt type."""

    def __bytes__(self):
        """Returns the respective bytes."""
        if self == 0:
            return self.to_bytes(1, 'little')

        value = int(self)
        bytes_ = b''

        while True:
            temp = value & 0b01111111
            value = rshift(value, 7)

            if value:
                temp |= 0b10000000

            bytes_ += temp.to_bytes(1, 'little')

            if not value:
                break

        return bytes_

    @classmethod
    def read(cls, readfunc: Callable) -> VarInt:
        """Reads a VarInt from the respective bytes."""
        bytes_count = 0
        result = 0

        while True:
            if bytes_count > 4:  # Compensate for start index 0.
                raise ValueError('VarInt is too big.')

            byte = readfunc(1)
            read = int.from_bytes(byte, 'little')
            value = read & 0b01111111
            shift = 7 * bytes_count
            result |= value << shift
            bytes_count += 1

            if (read & 0b10000000) == 0:
                break

        LOGGER.debug('Read %i bytes of VarInt.', bytes_count)
        return cls(result)



"""Common functions."""


__all__ = ['rshift']


def rshift(integer: int, shift: int) -> int:
    """Logical right binary shift."""

    if integer >= 0:
        return integer >> shift

    return (integer + 0x100000000) >> shift



"""Basic statistics protocol."""

from __future__ import annotations
from typing import NamedTuple

from mcipc.query.proto.common import MAGIC
from mcipc.query.proto.common import random_session_id
from mcipc.query.proto.common import ip_or_hostname
from mcipc.query.proto.common import IPAddressOrHostname
from mcipc.query.proto.common import Type


__all__ = ['Request', 'BasicStats', 'BasicStatsMixin']


class Request(NamedTuple):
    """Basic statistics request packet."""

    magic: bytes
    type: Type
    session_id: int
    challenge_token: int

    def __bytes__(self):
        """Returns the packet as bytes."""
        payload = self.magic
        payload += bytes(self.type)
        payload += self.session_id.to_bytes(4, 'big', signed=True)
        payload += self.challenge_token.to_bytes(4, 'big', signed=True)
        return payload

    @classmethod
    def create(cls, challenge_token: int, session_id: int = None) -> Request:
        """Creates a new request with the specified challenge
        token and the specified or a random session ID.
        """
        if session_id is None:
            session_id = random_session_id()

        return cls(MAGIC, Type.STAT, session_id, challenge_token)


class BasicStats(NamedTuple):
    """Basic statistics response packet."""

    type: Type
    session_id: int
    motd: str
    game_type: str
    map: str
    num_players: int
    max_players: int
    host_port: int
    host_ip: IPAddressOrHostname

    @classmethod
    def from_bytes(cls, bytes_: bytes) -> BasicStats:   # pylint: disable=R0914
        """Creates the packet from the respective bytes."""
        type_ = Type.from_bytes(bytes_[0:1])
        session_id = int.from_bytes(bytes_[1:5], 'big', signed=True)
        motd, *blocks, port_ip, _ = bytes_[5:].split(b'\0')
        motd = motd.decode('latin-1')
        strings = [block.decode('latin-1') for block in blocks]
        game_type, map_, num_players, max_players = strings
        num_players = int(num_players)
        max_players = int(max_players)
        host_port = int.from_bytes(port_ip[0:2], 'little')
        host_ip = ip_or_hostname(port_ip[2:].decode())
        return cls(
            type_, session_id, motd, game_type, map_, num_players, max_players,
            host_port, host_ip)

    def to_json(self) -> dict:
        """Returns a JSON-ish dict."""
        return {
            'type': self.type.value,
            'session_id': self.session_id,
            'motd': self.motd,
            'game_type': self.game_type,
            'map': self.map,
            'num_players': self.num_players,
            'max_players': self.max_players,
            'host_port': self.host_port,
            'host_ip': str(self.host_ip)
        }


class BasicStatsMixin:  # pylint: disable=R0903
    """Query client mixin for basic stats."""

    @property
    def basic_stats(self) -> BasicStats:
        """Returns basic stats"""
        request = Request.create(self._challenge_token)
        return self.communicate(request, BasicStats)



"""High level client API."""

from json import dumps
from typing import Tuple

from mcipc.rcon.datastructures import Help, Location, Players, Seed
from mcipc.rcon.exceptions import NotALocation, StructureNotFound
from mcipc.rcon.proto import Client as _Client


__all__ = ['Client']


Coords = Tuple[float, float, float]
Orientation = Tuple[float, float]


class AdminMixin:
    """Administrative methods."""

    def deop(self, player: str) -> str:
        """Revokes operator status from the respective player."""
        return self.run('deop', player)

    def kick(self, player: str, *reasons: str) -> str:
        """Kicks the respective player."""
        return self.run('kick', player, *reasons)

    def mkop(self, player: str) -> str:
        """Makes the respective player an operator."""
        return self.run('op', player)

    def teleport(self, player: str, *,
                 dst_player: str = None,
                 coords: Coords = None,
                 orientation: Orientation = None) -> str:
        """Teleports the player to either another
        player or a set of coordinates.
        """
        args = [player]

        if sum(item is None for item in (dst_player, coords)) != 1:
            raise TypeError('Must specify either dst_player or coords.')

        if dst_player is not None:
            args.append(dst_player)
        elif coords is not None:
            coord_x, coord_y, coord_z = coords
            args += [str(coord_x), str(coord_y), str(coord_z)]

        if orientation is not None:
            yaw, pitch = orientation
            args += [str(yaw), str(pitch)]

        return self.run('tp', *args)

    tp = teleport   # Alias.


class ChatMixin:
    """Mixin provinding chat-related methods."""

    def me(self, message: str) -> str:  # pylint: disable=C0103
        """Sends a message from RCON in first-person perspective."""
        return self.run('me', message)

    def tell(self, player: str, message: str) -> str:
        """Whispers a message to the respective player."""
        return self.run('tell', player, message)

    w = msg = tell  # Aliases.

    def say(self, message: str) -> str:
        """Broadcast a message to all players."""
        return self.run('say', message)

    def tellraw(self, player: str, obj: dict) -> str:
        """Sends a message represented by a JSON-ish dict."""
        return self.run('tellraw', player, dumps(obj))

    def send_url(self, player: str, url: str, text: str = None) -> str:
        """Sends a URL to the specified player.
        If text is None, it will default to the original URL.
        """
        json = {
            'text': text or url,
            'clickEvent': {
                'action': 'open_url',
                'value': url
            }
        }
        return self.tellraw(player, json)


class InfoMixin:
    """Server information mixin."""

    @property
    def players(self) -> Players:
        """Returns the players."""
        response = self.run('list')
        return Players.from_response(response)

    @property
    def seed(self) -> Seed:
        """Returns the server seed."""
        response = self.run('seed')
        return Seed.from_response(response)

    def help(self, command: str = None) -> Help:
        """Returns help about commands."""
        if command is None:
            text = self.run('help')
        else:
            text = self.run('help', command)

        items = filter(None, text.split('/'))
        command_args = (item.split(maxsplit=1) for item in items)
        return Help.from_sequence(command_args)

    def locate(self, structure: str) -> Location:
        """Locates the respective structure."""
        response = self.run('locate', structure)

        try:
            return Location.from_response(response)
        except NotALocation:
            raise StructureNotFound(structure) from None


class Client(_Client, AdminMixin, ChatMixin, InfoMixin):
    """A high-level RCON client."""



"""Query client CLI."""

from argparse import ArgumentParser, Namespace
from json import dumps
from logging import DEBUG, INFO, basicConfig, getLogger
from socket import timeout
from sys import exit, stdout    # pylint: disable=W0622
from typing import Tuple

from mcipc.config import LOG_FORMAT, InvalidCredentials, Credentials
from mcipc.query.client import Client
from mcipc.query.config import CONFIG


__all__ = ['main']


DEFAULT_INDENT = 2 if stdout.isatty() else None
LOGGER = getLogger('queryclt')


def get_args() -> Namespace:
    """Parses and returns the CLI arguments."""

    parser = ArgumentParser(description='A Minecraft Query client.')
    parser.add_argument('server', help='the server to connect to')
    parser.add_argument(
        '-i', '--indent', type=int, default=DEFAULT_INDENT,
        help='indentation for JSON output')
    parser.add_argument(
        '-t', '--timeout', type=float, help='connection timeout in seconds')
    parser.add_argument(
        '-d', '--debug', action='store_true',
        help='print additional debug information')
    subparsers = parser.add_subparsers(dest='action')
    # Basic stats.
    basic_stats_parser = subparsers.add_parser(
        'basic-stats', help='returns basic stats')
    basic_stats_field = basic_stats_parser.add_subparsers(dest='field')
    basic_stats_field.add_parser('motd', help='returns the message of the day')
    basic_stats_field.add_parser('game-type', help='returns the game type')
    basic_stats_field.add_parser('map', help='returns the map name')
    basic_stats_field.add_parser(
        'num-players', help='returns the amount of online players')
    basic_stats_field.add_parser(
        'max-players', help='returns the maximally allowed players')
    basic_stats_field.add_parser(
        'host-port', help='returns the port of the server')
    basic_stats_field.add_parser(
        'host-ip', help='returns the IP address of the server')
    # Full stats.
    full_stats_parser = subparsers.add_parser(
        'full-stats', help='returns full stats')
    full_stats_field = full_stats_parser.add_subparsers(dest='field')
    full_stats_field.add_parser(
        'hostname', help='returns the host name of the server (same as motd)')
    full_stats_field.add_parser('game-type', help='returns the game type')
    full_stats_field.add_parser('game-id', help='returns the game ID')
    full_stats_field.add_parser('version', help='returns the game version')
    full_stats_field.add_parser('plugins', help='returns the plugins in use')
    full_stats_field.add_parser('map', help='returns the map name')
    full_stats_field.add_parser(
        'num-players', help='returns the amount of online players')
    full_stats_field.add_parser(
        'max-players', help='returns the maximally allowed players')
    full_stats_field.add_parser(
        'host-port', help='returns the port of the server')
    full_stats_field.add_parser(
        'host-ip', help='returns the IP address of the server')
    full_stats_field.add_parser(
        'players', help='returns a list of online players')
    return parser.parse_args()


def get_credentials(server: str) -> Tuple[str, int]:
    """Get the credentials for a server from the respective server name."""

    try:
        host, port, passwd = Credentials.from_string(server)
    except InvalidCredentials:
        try:
            host, port, passwd = CONFIG.servers[server]
        except KeyError:
            LOGGER.error('No such server: %s.', server)
            exit(2)

    if passwd is not None:
        LOGGER.warning('Query protocol does not require a password.')

    return (host, port)


def basic_stats(client: Client, args: Namespace):   # pylint: disable=R0911
    """Handles basic stats queries."""

    if not args.field:
        return dumps(client.basic_stats.to_json(), indent=args.indent)

    if args.field == 'motd':
        return client.basic_stats.motd

    if args.field == 'game-type':
        return client.basic_stats.game_type

    if args.field == 'map':
        return client.basic_stats.map

    if args.field == 'num-players':
        return client.basic_stats.num_players

    if args.field == 'max-players':
        return client.basic_stats.max_players

    if args.field == 'host-port':
        return client.basic_stats.host_port

    if args.field == 'host-ip':
        return client.basic_stats.host_ip

    raise ValueError('Invalid action.')


def full_stats(client: Client, args: Namespace):    # pylint: disable=R0911
    """Handles full stats queries."""

    if not args.field:
        return dumps(client.full_stats.to_json(), indent=args.indent)

    if args.field == 'hostname':
        return client.full_stats.host_name

    if args.field == 'game-type':
        return client.full_stats.game_type

    if args.field == 'game-id':
        return client.full_stats.game_id

    if args.field == 'version':
        return client.full_stats.version

    if args.field == 'plugins':
        return dumps(client.full_stats.plugins, indent=args.indent)

    if args.field == 'map':
        return client.full_stats.map

    if args.field == 'num-players':
        return client.full_stats.num_players

    if args.field == 'max-players':
        return client.full_stats.max_players

    if args.field == 'host-port':
        return client.full_stats.host_port

    if args.field == 'host-ip':
        return client.full_stats.host_ip

    if args.field == 'players':
        return dumps(client.full_stats.players, indent=args.indent)

    raise ValueError('Invalid action.')


def main():
    """Runs the RCON client."""

    args = get_args()
    log_level = DEBUG if args.debug else INFO
    basicConfig(level=log_level, format=LOG_FORMAT)
    host, port = get_credentials(args.server)

    try:
        with Client(host, port, timeout=args.timeout) as client:
            if args.action == 'basic-stats':
                print(basic_stats(client, args), flush=True)
            elif args.action == 'full-stats':
                print(full_stats(client, args), flush=True)
    except timeout:
        LOGGER.error('Connection timeout.')
        exit(3)



"""Locations."""

from __future__ import annotations
from re import compile  # pylint: disable=W0622
from typing import NamedTuple, Union

from mcipc.rcon.exceptions import NotALocation


__all__ = ['Location']


REGEX = compile('.*\\[(-?\\d+), (~|-?\\d+), (-?\\d+)\\].*')


def _int_or_none(string: str) -> Union[int, None]:
    """Returns None iff coordinate is special "~"
    character or else the respective integer value.
    """

    if string == '~':
        return None

    return int(string)


class Location(NamedTuple):
    """A 3D location."""

    x: int
    y: Union[int, None]
    z: int

    @classmethod
    def from_response(cls, text: str) -> Location:
        """Creates a location from a server response."""
        match = REGEX.fullmatch(text)

        if match is None:
            raise NotALocation(text)

        coord_x, coord_y, coord_z = match.groups()
        return cls(int(coord_x), _int_or_none(coord_y), int(coord_z))

    def to_json(self) -> dict:
        """Returns a JSON-ish dict."""
        return {'x': self.x, 'y': self.y, 'z': self.z}



"""Low-level protocol stuff."""

from __future__ import annotations
from enum import Enum
from logging import getLogger
from random import randint
from socket import SOCK_STREAM
from typing import NamedTuple

from mcipc.common import BaseClient
from mcipc.rcon.exceptions import InvalidPacketStructure
from mcipc.rcon.exceptions import RequestIdMismatch
from mcipc.rcon.exceptions import WrongPassword


__all__ = ['Type', 'Packet', 'Client']


LOGGER = getLogger(__file__)
TAIL = b'\0\0'


def random_request_id() -> int:
    """Returns a random, positive int32."""

    return randint(0, 2147483647 + 1)


class Type(Enum):
    """Available packet types."""

    LOGIN = 3
    COMMAND = 2
    RESPONSE = 0

    def __int__(self):
        """Returns the actual integer value."""
        return self.value

    def __bytes__(self):
        """Returns the integer value as little endian."""
        return int(self).to_bytes(4, 'little')


class Packet(NamedTuple):
    """An RCON packet."""

    request_id: int
    type: Type
    payload: str

    def __bytes__(self):
        """Returns the packet as bytes."""
        payload = self.request_id.to_bytes(4, 'little', signed=True)
        payload += bytes(self.type)
        payload += self.payload.encode()
        payload += TAIL
        size = len(payload).to_bytes(4, 'little')
        return size + payload

    @classmethod
    def from_bytes(cls, bytes_: bytes) -> Packet:
        """Creates a packet from the respective bytes."""
        request_id = int.from_bytes(bytes_[:4], 'little', signed=True)
        type_ = int.from_bytes(bytes_[4:8], 'little')
        payload = bytes_[8:-2]
        tail = bytes_[-2:]

        if tail != TAIL:
            raise InvalidPacketStructure('Invalid tail.', tail, TAIL)

        return cls(request_id, Type(type_), payload.decode())

    @classmethod
    def from_args(cls, *args: str) -> Packet:
        """Creates a command packet."""
        return cls(random_request_id(), Type.COMMAND, ' '.join(args))

    @classmethod
    def from_login(cls, passwd: str) -> Packet:
        """Creates a login packet."""
        return cls(random_request_id(), Type.LOGIN, passwd)


class Client(BaseClient):
    """An RCON client."""

    def __init__(self, host: str, port: int, timeout: float = None):
        """Initializes the base client with the SOCK_STREAM socket type."""
        super().__init__(SOCK_STREAM, host, port, timeout=timeout)

    def communicate(self, packet: Packet) -> Packet:
        """Sends and receives a packet."""
        self._socket.send(bytes(packet))
        header = self._socket.recv(4)
        length = int.from_bytes(header, 'little')
        payload = self._socket.recv(length)
        response = Packet.from_bytes(payload)

        if response.request_id == packet.request_id:
            return response

        raise RequestIdMismatch(packet.request_id, response.request_id)

    def login(self, passwd: str) -> bool:
        """Performs a login."""
        packet = Packet.from_login(passwd)

        try:
            self.communicate(packet)
        except RequestIdMismatch:
            raise WrongPassword() from None

        return True

    def run(self, command: str, *arguments: str, raw: bool = False) -> str:
        """Runs a command."""
        packet = Packet.from_args(command, *arguments)
        response = self.communicate(packet)
        return response if raw else response.payload



"""Full statistics protocol."""

from __future__ import annotations
from typing import Generator, NamedTuple, Tuple

from mcipc.query.proto.common import MAGIC
from mcipc.query.proto.common import random_session_id
from mcipc.query.proto.common import ip_or_hostname
from mcipc.query.proto.common import IPAddressOrHostname
from mcipc.query.proto.common import Type


__all__ = ['Request', 'FullStats', 'FullStatsMixin']


PADDING = 0
NULL = b'\0'


def get_dict(bytes_: bytes) -> Tuple[int, dict]:
    """Returns the end index and a dictionary
    of zero-separated key-value pairs.
    """

    item = ''
    dictionary = {}
    key = None
    is_key = True

    for index, integer in enumerate(bytes_):
        byte = bytes([integer])

        if byte == NULL:
            if not item and is_key:
                return (index, dictionary)

            if is_key:
                key = item
                is_key = False
            else:
                dictionary[key] = item
                key = None
                is_key = True

            item = ''
        else:
            item += byte.decode('latin-1')

    raise ValueError('Bytes string not properly terminated.', bytes_)


def items(bytes_: bytes) -> Generator[str, None, None]:
    """Yields zero-byte-separated items."""

    item = ''

    for integer in bytes_:
        byte = bytes([integer])

        if byte == NULL:
            if not item:
                return

            yield item
            item = ''
        else:
            item += byte.decode('latin-1')


def plugins_to_dict(string: str) -> dict:
    """Convers a plugins string into a dictionary."""

    try:
        mod, plugins = string.split(': ')
    except ValueError:  # No plugins.
        return {}

    return {mod: plugins.split('; ')}


def stats_from_dict(dictionary: dict):
    """Yields statistics options from the provided dictionary."""

    yield dictionary['hostname']
    yield dictionary['gametype']
    yield dictionary['game_id']
    yield dictionary['version']
    yield plugins_to_dict(dictionary['plugins'])
    yield dictionary['map']
    yield int(dictionary['numplayers'])
    yield int(dictionary['maxplayers'])
    yield int(dictionary['hostport'])
    yield ip_or_hostname(dictionary['hostip'])


class Request(NamedTuple):
    """Basic statistics request packet."""

    magic: bytes
    type: Type
    session_id: int
    challenge_token: int
    padding: int

    def __bytes__(self):
        """Returns the packet as bytes."""
        payload = self.magic
        payload += bytes(self.type)
        payload += self.session_id.to_bytes(4, 'big', signed=True)
        payload += self.challenge_token.to_bytes(4, 'big', signed=True)
        payload += self.padding.to_bytes(4, 'big')
        return payload

    @classmethod
    def create(cls, challenge_token: int, session_id: int = None) -> Request:
        """Creates a new request with the specified challenge
        token and the specified or a random session ID.
        """
        if session_id is None:
            session_id = random_session_id()

        return cls(MAGIC, Type.STAT, session_id, challenge_token, PADDING)


class FullStats(NamedTuple):
    """Full statistics response."""

    type: Type
    session_id: int
    host_name: str
    game_type: str
    game_id: str
    version: str
    plugins: dict
    map: str
    num_players: int
    max_players: int
    host_port: int
    host_ip: IPAddressOrHostname
    players: tuple

    @classmethod
    def from_bytes(cls, bytes_: bytes) -> FullStats:
        """Creates the full stats object from the respective bytes."""
        type_ = Type.from_bytes(bytes_[0:1])
        session_id = int.from_bytes(bytes_[1:5], 'big', signed=True)
        index = 16  # Discard padding.
        index, stats = get_dict(bytes_[index:])
        index += 16 + 1     # Discard additional null byte.
        index += 10     # Discard padding.
        players = tuple(items(bytes_[index:]))
        return cls(type_, session_id, *stats_from_dict(stats), players)

    def to_json(self) -> dict:
        """Returns a JSON-ish dict."""
        return {
            'type': self.type.value,
            'session_id': self.session_id,
            'host_name': self.host_name,
            'game_type': self.game_type,
            'game_id': self.game_id,
            'version': self.version,
            'plugins': self.plugins,
            'map': self.map,
            'num_players': self.num_players,
            'max_players': self.max_players,
            'host_port': self.host_port,
            'host_ip': str(self.host_ip),
            'players': self.players
        }


class FullStatsMixin:   # pylint: disable=R0903
    """Query client mixin for full stats."""

    @property
    def full_stats(self) -> FullStats:
        """Returns full stats"""
        request = Request.create(self._challenge_token)
        return self.communicate(request, FullStats)



"""Query client library."""

from socket import SOCK_DGRAM
from typing import Generator

from mcipc.common import BaseClient
from mcipc.query.proto import BasicStatsMixin
from mcipc.query.proto import FullStatsMixin
from mcipc.query.proto import HandshakeMixin
from mcipc.query.proto import Packet


__all__ = ['Client']


class Client(BaseClient, HandshakeMixin, BasicStatsMixin, FullStatsMixin):
    """A Query client."""

    def __init__(self, host: str, port: int, timeout: float = None):
        """Initializes the base client with the socket
        type SOCK_DGRAM and sets the challenge token.
        """
        super().__init__(SOCK_DGRAM, host, port, timeout=timeout)
        self._challenge_token = None

    def __enter__(self):
        """Performs a handshake."""
        super().__enter__()

        if self._challenge_token is None:
            self._challenge_token = self.handshake().challenge_token

        return self

    def _recv_chunks(self, buffer: int = 4096) -> Generator[bytes, None, None]:
        """Yields chunks of bytes from the socket."""
        while chunk := self._socket.recv(buffer):
            yield chunk

            if len(chunk) < buffer:
                return

    def _recv_all(self, buffer: int = 4096) -> bytes:
        """Receives all bytes from the socket."""
        return b''.join(self._recv_chunks(buffer=buffer))

    def communicate(self, packet: Packet, response_type: type = None, *,
                    buffer: int = 4096):
        """Sends and receives a packet."""
        self._socket.send(bytes(packet))
        response = self._recv_all(buffer=buffer)

        if response_type is None:
            return response

        return response_type.from_bytes(response)



"""Experimental features that are subject to unannounced change."""

from datetime import datetime
from locale import LC_TIME, getdefaultlocale, setlocale
from subprocess import PIPE, check_output

from mcipc.config import FORTUNE
from mcipc.rcon import client


__all__ = ['Client']


class Client(client.Client):
    """Client with some more extras."""

    def fortune(self, short: bool = True, offensive: bool = False) -> str:
        """Sends a fortune to all players."""
        command = [FORTUNE]

        if short:
            command.append('-s')

        if offensive:
            command.append('-o')

        # pylint: disable=E1123
        text = check_output(command, stderr=PIPE, text=True)
        return self.say(text)

    def datetime(self, frmt: str = '%c') -> str:
        """Tells all players the current datetime."""
        setlocale(LC_TIME, getdefaultlocale())  # Fix loacale.
        text = datetime.now().strftime(frmt)
        return self.say(text)



"""The actual server."""

from json import dumps
from logging import getLogger
from socket import socket

from mcipc.server.datastructures import Handshake, SLPResponse
from mcipc.server.datatypes import VarInt
from mcipc.server.enumerations import State


__all__ = ['get_response', 'StubServer']


LOGGER = getLogger(__file__)


def get_response(text: str) -> bytes:
    """Returns the response text message."""

    payload = dumps({'text': text}).encode('latin-1')
    payload_length = bytes(VarInt(len(payload)))
    payload = bytes(VarInt(0)) + payload_length + payload
    payload_length = bytes(VarInt(len(payload)))
    return payload_length + payload


class StubServer:
    """A stub minecraft server."""

    def __init__(self, description: str, *, version: str = '1.14.2',
                 max_players: int = 20, protocol: int = 485):
        """Description, max players and protocol information."""
        self.description = description
        self.version = version
        self.max_players = max_players
        self.protocol = protocol

    @property
    def slp_response(self) -> SLPResponse:
        """Returns an SLP response."""
        json = {
            'version': {
                'name': self.version,
                'protocol': self.protocol
            },
            'players': {
                'max': self.max_players,
                'online': 0
            },
            'description': {
                'text': self.description
            }
        }
        return SLPResponse(VarInt(0), json)

    @staticmethod
    def _perform_handshake(connection: socket) -> State:
        """Handle handshake requests."""
        handshake = Handshake.read(connection.recv)
        LOGGER.debug('Got handshake: %s', handshake)
        return handshake.next_state

    def _perform_status(self, connection: socket):
        """Handles status requests."""
        packet_id = connection.recv(1)

        if packet_id == b'\x01':
            LOGGER.debug('Got packet id: %s', packet_id)
            slp_response = bytes(self.slp_response)
            LOGGER.debug('Sending SLP response: %s', slp_response)
            connection.send(slp_response)

    def _perform_login(self, connection: socket):
        """Handles the login response."""
        raise NotImplementedError()

    def _handle_login(self, connection: socket):
        """Performs a login."""
        size = VarInt.read(connection.recv)
        packet_id = VarInt.read(connection.recv)
        packet_id_length = len(bytes(packet_id))
        payload = connection.recv(size - packet_id_length)
        LOGGER.debug('Got packet ID: %s', packet_id)
        user_name = payload[2:].decode('latin-1')
        LOGGER.debug('User "%s" logged in.', user_name)
        self._perform_login(connection)

    def _process(self, connection: socket, state: State = State.HANDSHAKE):
        """Runs the connection processing."""
        LOGGER.debug('Current state: %s', state)

        if state == State.HANDSHAKE:
            state = self._perform_handshake(connection)
            self._process(connection, state=state)
        elif state == State.STATUS:
            self._perform_status(connection)
        elif state == State.LOGIN:
            self._handle_login(connection)

    def spawn(self, address: str, port: int):
        """Spawns the server on the respective socket."""
        with socket() as sock:
            sock.bind((address, port))
            sock.listen()

            while True:
                connection, address = sock.accept()

                with connection:
                    LOGGER.debug('New connection from: %s', address)
                    self._process(connection)



"""Datastructures to represent help on commands."""

from __future__ import annotations
from logging import getLogger
from typing import Iterable, NamedTuple, Tuple, Union


LOGGER = getLogger(__file__)


class Command(NamedTuple):
    """Represents a command / arguments tuple."""

    command: str
    arguments: str

    @classmethod
    def from_tuple(cls, tpl: Union[Tuple[str, str], Tuple[str]]) -> Command:
        """Creates the command from a tuple."""
        try:
            command, arguments, *superfluous = tpl
        except ValueError:
            command, = tpl
            arguments = None
            superfluous = None

        if superfluous:
            items = len(superfluous) + 2
            raise ValueError(f'Expected one or two items. Got {items}.')

        return cls(command, arguments)

    @property
    def usage(self) -> str:
        """Returns a docopt-compliant usage string."""
        usage = f'Usage: {self.command}'

        if self.arguments:
            usage += f' {self.arguments}'

        return usage

    def to_json(self) -> dict:
        """Returns a JSON-ish dict."""
        return {'command': self.command, 'arguments': self.arguments}


class Help(dict):
    """command: arguments key pairs."""

    @classmethod
    def from_sequence(cls, sequence: Iterable) -> Help:
        """Creates the help from the respective sequence."""
        commands = (Command.from_tuple(item) for item in sequence)
        return cls((command.command, command) for command in commands)

    def to_json(self) -> dict:
        """Returns a JSON-ish dict."""
        return {name: command.to_json() for name, command in self.items()}



#! /usr/bin/env python
"""Installation script."""

from setuptools import setup

setup(
    name='mcipc',
    version_format='{tag}',
    setup_requires=['setuptools-git-version'],
    author='Richard Neumann',
    author_email='mail@richard-neumann.de',
    python_requires='>=3.8',
    packages=[
        'mcipc',
        'mcipc.cli',
        'mcipc.query',
        'mcipc.query.proto',
        'mcipc.rcon',
        'mcipc.rcon.datastructures',
        'mcipc.server'
    ],
    scripts=[
        'files/mcstubsrv',
        'files/queryclt',
        'files/rconclt',
        'files/rconshell'
    ],
    url='https://github.com/conqp/mcipc',
    license='GPLv3',
    description='A Minecraft server inter-process communication library.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    keywords='minecraft python server rcon query'
)



"""Common protocol stuff."""

from __future__ import annotations
from enum import Enum
from ipaddress import ip_address, IPv4Address, IPv6Address
from random import randint
from typing import Union


__all__ = [
    'MAGIC',
    'random_session_id',
    'ip_or_hostname',
    'IPAddressOrHostname',
    'Type'
]


MAGIC = b'\xfe\xfd'
SESSION_ID_MASK = 0x0F0F0F0F
IPAddressOrHostname = Union[IPv4Address, IPv6Address, str]


def random_session_id() -> int:
    """Returns a random session ID.
    See: https://wiki.vg/Query#Generating_a_Session_ID
    """

    return randint(-2147483648, 2147483647 + 1) & SESSION_ID_MASK


def ip_or_hostname(string: str) -> IPAddressOrHostname:
    """Returns an IPv4 or IPv6 address if applicable, else a string."""

    try:
        return ip_address(string)
    except ValueError:
        return string


class Type(Enum):
    """Request type."""

    HANDSHAKE = 9
    STAT = 0

    def __bytes__(self):
        """Reuturns the integer value as big endian."""
        return self.value.to_bytes(1, 'big')    # pylint: disable=E1101

    @classmethod
    def from_bytes(cls, bytes_: bytes) -> Type:
        """Returns the type from the respective bytes."""
        return cls(int.from_bytes(bytes_, 'big'))



"""RCON exceptions."""


__all__ = [
    'InvalidPacketStructure',
    'NotALocation',
    'RequestIdMismatch',
    'StructureNotFound',
    'WrongPassword'
]


class InvalidPacketStructure(ValueError):
    """Indicates an invalid packet structure."""

    def __init__(self, message: str, value, expected):
        """Sets an error message, the actual value and the expected value."""
        super().__init__(message, value, expected)
        self.message = message
        self.value = value
        self.expected = expected

    @property
    def description(self) -> str:
        """Returns an additional error description."""
        return f'Expected "{self.expected}", but got "{self.value}".'

    def __str__(self):
        """Returns a string representation."""
        return f'{self.message} {self.description}'


class NotALocation(ValueError):
    """Indicates that the given text is not a valid location value."""


class RequestIdMismatch(Exception):
    """Indicates that the sent and received request IDs do not match."""

    def __init__(self, sent: int, received: int):
        """Sets the sent and received request IDs."""
        super().__init__(sent, received)
        self.sent = sent
        self.received = received


class StructureNotFound(ValueError):
    """Indicates that the given structure could not be found."""

    def __init__(self, structure):
        """Sets the invalid structure's name."""
        super().__init__(structure)
        self.structure = structure


class WrongPassword(Exception):
    """Indicates a wrong RCON password."""



"""Stub server launcher."""

from argparse import ArgumentParser, Namespace
from logging import getLogger
from socket import socket
from subprocess import CalledProcessError, check_call
from sys import exit    # pylint: disable=W0622

from mcipc.server import get_response, StubServer


__all__ = ['main']


DESCRIPTION = 'Starts a stub server'
LOGGER = getLogger(__file__)
DEFAULT_PLAYERS = 20
DEFAULT_PROTO = 485
DEFAULT_UNIT_TEMPLATE = 'minecraft@{}.server'


def get_args() -> Namespace:
    """Parses the command line arguments."""

    parser = ArgumentParser(description=DESCRIPTION)
    parser.add_argument('socket', help='the socket to listen on')
    parser.add_argument('name', help='the server name')
    parser.add_argument('description', help='server description')
    parser.add_argument('-p', '--players', type=int, default=DEFAULT_PLAYERS,
                        help='amount of player slots')
    parser.add_argument('-P', '--protocol', type=int, default=DEFAULT_PROTO,
                        help='protocol version')
    parser.add_argument('-t', '--template', default=DEFAULT_UNIT_TEMPLATE,
                        help='systemd unit template')
    return parser.parse_args()


class ServerLauncher(StubServer):
    """Server that launches the actual server."""

    def __init__(self, name: str, description: str,     # pylint: disable=R0913
                 max_players: int = DEFAULT_PLAYERS,
                 protocol: int = DEFAULT_PROTO,
                 template: str = DEFAULT_UNIT_TEMPLATE):
        """Sets server meta data."""
        super().__init__(
            description, max_players=max_players, protocol=protocol)
        self.name = name
        self.template = template

    @property
    def unit(self) -> str:
        """Returns the system unit to start."""
        return self.template.format(self.name)

    def _start_server(self) -> bool:
        """Starts the server."""
        command = ('/bin/systemctl', 'start', self.unit)

        try:
            check_call(command)
        except CalledProcessError as cpe:
            LOGGER.debug(cpe)
            return False

        return True

    def _perform_login(self, connection: socket):
        """Performs a login."""
        if self._start_server():
            response = get_response('Server has been started.')
        else:
            response = get_response('Server could not be started.')

        connection.send(response)


def main():
    """Main program."""

    args = get_args()

    try:
        host, port = args.socket.split(':')
    except ValueError:
        LOGGER.error('Invalid socket: %s', args.socket)
        exit(1)

    try:
        port = int(port)
    except ValueError:
        LOGGER.error('Invalid port: %s', port)
        exit(2)

    server = ServerLauncher(
        args.name, args.description, max_players=args.players,
        protocol=args.protocol, template=args.template)

    try:
        server.spawn(host, port)
    except OSError as error:
        LOGGER.error(str(error))
        exit(3)
    except KeyboardInterrupt:
        LOGGER.info('Shutting down...')



"""An IPC library to communicate with Minecraft servers."""

from mcipc.query import Client as QueryClient
from mcipc.rcon import Client as RCONClient
from mcipc.rcon import InvalidPacketStructure
from mcipc.rcon import StructureNotFound
from mcipc.rcon import WrongPassword
from mcipc.rcon import rconcmd


__all__ = [
    'InvalidPacketStructure',
    'StructureNotFound',
    'WrongPassword',
    'rconcmd',
    'QueryClient',
    'RCONClient'
]



"""Minecraft server protocol."""

from mcipc.server.datastructures import Handshake, SLPResponse
from mcipc.server.datatypes import VarInt
from mcipc.server.enumerations import State
from mcipc.server.server import get_response, StubServer


__all__ = [
    'get_response',
    'Handshake',
    'SLPResponse',
    'State',
    'StubServer',
    'VarInt'
]



"""RCON client CLI."""

from argparse import ArgumentParser, Namespace
from getpass import getpass
from logging import DEBUG, INFO, basicConfig, getLogger
from socket import timeout
from subprocess import CalledProcessError, check_call
from sys import exit    # pylint: disable=W0622
from typing import Tuple

from mcipc.config import LOG_FORMAT, InvalidCredentials, Credentials
from mcipc.rcon.config import CONFIG
from mcipc.rcon.datastructures import Players
from mcipc.rcon.playground import Client


__all__ = ['get_credentials', 'main']


LOGGER = getLogger('rconclt')


def get_args() -> Namespace:
    """Parses and returns the CLI arguments."""

    parser = ArgumentParser(description='A Minecraft RCON client.')
    parser.add_argument('server', help='the server to connect to')
    parser.add_argument(
        '-t', '--timeout', type=float, help='connection timeout in seconds')
    parser.add_argument(
        '-d', '--debug', action='store_true',
        help='print additional debug information')
    subparsers = parser.add_subparsers(dest='action')
    command_parser = subparsers.add_parser(
        'exec', help='execute commands on the server')
    command_parser.add_argument(
        'command', help='command to execute on the server')
    command_parser.add_argument(
        'argument', nargs='*', default=(), help='arguments for the command')
    say_parser = subparsers.add_parser(
        'say', help='broadcast a message on the server')
    say_parser.add_argument('message', help='the message to broadcast')
    fortune_parser = subparsers.add_parser(
        'fortune', help='send a fortune to the players on the server')
    fortune_parser.add_argument(
        '-l', '--long', action='store_true', help='generate ling fortunes')
    fortune_parser.add_argument(
        '-o', '--offensive', action='store_true',
        help='generate offensive fortunes')
    datetime_parser = subparsers.add_parser(
        'datetime',
        help='sends the current date and time to the players on the server')
    datetime_parser.add_argument(
        '-f', '--format', default='%c', help='the datetime format')
    subparsers.add_parser('in-use', help='checks whether the server is in use')
    shutdown_parser = subparsers.add_parser(
        'idle-shutdown', help='shuts down the server if it is not in use')
    shutdown_parser.add_argument(
        '-s', '--sudo', action='store_true',
        help='invoke the shutdown command using sudo')
    shutdown_parser.add_argument(
        '-u', '--unit', default='minecraft@{server}.service',
        help='the systemd unit template')
    return parser.parse_args()


def get_credentials(server: str) -> Tuple[str, int, str]:
    """Get the credentials for a server from the respective server name."""

    try:
        host, port, passwd = Credentials.from_string(server)
    except InvalidCredentials:
        try:
            host, port, passwd = CONFIG.servers[server]
        except KeyError:
            LOGGER.error('No such server: %s.', server)
            exit(2)

    if passwd is None:
        try:
            passwd = getpass('Password: ')
        except (KeyboardInterrupt, EOFError):
            print()
            LOGGER.error('Aborted by user.')
            exit(3)

    return (host, port, passwd)


def idle_shutdown(players: Players, args: Namespace) -> bool:
    """Shuts down the server if it is idle."""

    if players.online:
        LOGGER.info('Server is in use.')
        return False

    LOGGER.info('Server is idle.')
    unit = args.unit.format(server=args.server)
    command = ('/usr/bin/systemctl', 'stop', unit)

    try:
        check_call(command)
    except CalledProcessError as error:
        LOGGER.error('Could not shutdown the server.')
        LOGGER.debug(error)
        return False

    LOGGER.info('Server %s has been shut down.', unit)
    return True


def run_action(client: Client, args: Namespace):
    """Runs the respective actions."""

    result = None

    if args.action == 'exec':
        result = client.run(args.command, *args.argument)
    elif args.action == 'say':
        result = client.say(args.message)
    elif args.action == 'fortune':
        result = client.fortune(
            short=not args.long, offensive=args.offensive)
    elif args.action == 'datetime':
        result = client.datetime(frmt=args.format)
    elif args.action == 'in-use':
        players = client.players

        if players.online:
            LOGGER.info('There are %i players online:', players.online)
            LOGGER.info(', '.join(players.names))
        else:
            LOGGER.warning('There are no players online.')
            exit(1)

    if result:
        LOGGER.info(result)


def main():
    """Runs the RCON client."""

    args = get_args()
    log_level = DEBUG if args.debug else INFO
    basicConfig(level=log_level, format=LOG_FORMAT)
    host, port, passwd = get_credentials(args.server)

    try:
        with Client(host, port, timeout=args.timeout) as client:
            if not client.login(passwd):
                LOGGER.error('Failed to log in.')
                exit(4)

            if args.action == 'idle-shutdown':
                players = client.players
            else:
                run_action(client, args)
    except timeout:
        LOGGER.error('Connection timeout.')
        exit(3)

    if args.action == 'idle-shutdown':
        if not idle_shutdown(players, args):
            exit(1)
